@@ -11,6 +11,7 @@
 #include <sys/ioctl.h>
 #include <sys/socket.h>
 #include <sys/types.h>
+#include <sys/un.h>
 #include <sys/wait.h>
 #include <linux/if_tun.h>
 #include <arpa/inet.h>
@@ -19,7 +20,10 @@
 #include <getopt.h>
 #include <stdbool.h>
 #include <regex.h>
+#include <libslirp.h>
 #include "slirp4netns.h"
+#include <ifaddrs.h>
+#include <seccomp.h>
 
 #define DEFAULT_MTU (1500)
 #define DEFAULT_CIDR ("10.0.2.0/24")
@@ -28,6 +32,7 @@
 #define DEFAULT_VNAMESERVER_OFFSET (3) // 10.0.2.3
 #define DEFAULT_RECOMMENDED_VGUEST_OFFSET (100) // 10.0.2.100
 #define DEFAULT_NETNS_TYPE ("pid")
+#define DEFAULT_TARGET_TYPE ("netns")
 #define NETWORK_PREFIX_MIN (1)
 // >=26 is not supported because the recommended guest IP is set to network addr
 // + 100 .
@@ -78,6 +83,10 @@ static int open_tap(const char *tapname)
 {
     int fd;
     struct ifreq ifr;
+    if (tapname == NULL) {
+        fprintf(stderr, "tapname is NULL\n");
+        return -1;
+    }
     if ((fd = open("/dev/net/tun", O_RDWR)) < 0) {
         perror("open(\"/dev/net/tun\")");
         return fd;
@@ -157,6 +166,14 @@ static int configure_network(const char *tapname,
         return -1;
     }
 
+    if (cfg->vmacaddress_len > 0) {
+        ifr.ifr_ifru.ifru_hwaddr = cfg->vmacaddress;
+        if (ioctl(sockfd, SIOCSIFHWADDR, &ifr) < 0) {
+            perror("cannot set MAC address");
+            return -1;
+        }
+    }
+
     sai->sin_family = AF_INET;
     sai->sin_port = 0;
     sai->sin_addr = cfg->recommended_vguest;
@@ -194,6 +211,7 @@ static int configure_network(const char *tapname,
     return 0;
 }
 
+/* Child (--target-type=netns) */
 static int child(int sock, pid_t target_pid, bool do_config_network,
                  const char *tapname, char *netns_path, char *userns_path,
                  struct slirp4netns_config *cfg)
@@ -218,6 +236,70 @@ static int child(int sock, pid_t target_pid, bool do_config_network,
     return 0;
 }
 
+/*
+ * Child (--target-type=bess)
+ *
+ * FIXME: We do not really need to fork the child for BESS mode
+ */
+static int child_bess(int sock, const char *bess_socket)
+{
+    int bess_listenfd = -1, bess_acceptfd = -1;
+    struct sockaddr_un bess_listenaddr, bess_acceptaddr;
+    socklen_t bess_acceptaddrlen = sizeof(struct sockaddr_un);
+    memset(&bess_acceptaddr, 0, sizeof(bess_acceptaddr));
+    memset(&bess_listenaddr, 0, sizeof(bess_listenaddr));
+
+    /* Listen on the BESS socket */
+    if ((bess_listenfd = socket(AF_UNIX, SOCK_SEQPACKET, 0)) == -1) {
+        perror("child_bess: socket");
+        goto error;
+    }
+    bess_listenaddr.sun_family = AF_UNIX;
+    if (bess_socket == NULL) {
+        fprintf(stderr, "the specified BESS socket path is NULL\n");
+        goto error;
+    }
+    if (strlen(bess_socket) >= sizeof(bess_listenaddr.sun_path)) {
+        fprintf(stderr, "the specified BESS socket path is too long (>= %lu)\n",
+                sizeof(bess_listenaddr.sun_path));
+        goto error;
+    }
+    strncpy(bess_listenaddr.sun_path, bess_socket,
+            sizeof(bess_listenaddr.sun_path) - 1);
+    unlink(bess_socket); /* avoid EADDRINUSE */
+    if (bind(bess_listenfd, (struct sockaddr *)&bess_listenaddr,
+             sizeof(bess_listenaddr)) < 0) {
+        perror("child_bess: bind");
+        goto error;
+    }
+    if (listen(bess_listenfd, 0) < 0) {
+        perror("child_bess: listen");
+        goto error;
+    }
+
+    /* Accept a connection, and send its connection to the parent */
+    fprintf(stderr, "Waiting for connection to %s\n", bess_socket);
+    if ((bess_acceptfd =
+             accept(bess_listenfd, (struct sockaddr *)&bess_acceptaddr,
+                    &bess_acceptaddrlen)) < 0) {
+        perror("child_bess: accept");
+        goto error;
+    }
+    if (sendfd(sock, bess_acceptfd) < 0) {
+        goto error;
+    }
+    fprintf(stderr, "sent \"tapfd\"=%d (not really TAP) for %s\n",
+            bess_acceptfd, bess_socket);
+    close(sock);
+    close(bess_listenfd);
+    return 0;
+error:
+    close(bess_acceptfd);
+    close(bess_listenfd);
+    close(sock);
+    return -1;
+}
+
 static int recvfd(int sock)
 {
     int fd;
@@ -254,7 +336,15 @@ static int recvfd(int sock)
 static int parent(int sock, int ready_fd, int exit_fd, const char *api_socket,
                   struct slirp4netns_config *cfg, pid_t target_pid)
 {
+    char str[INET6_ADDRSTRLEN];
     int rc, tapfd;
+    struct in_addr vdhcp_end = {
+#define NB_BOOTP_CLIENTS 16
+        /* NB_BOOTP_CLIENTS is hard-coded to 16 in libslirp:
+           https://gitlab.freedesktop.org/slirp/libslirp/-/issues/49 */
+        .s_addr = htonl(ntohl(cfg->vdhcp_start.s_addr) + NB_BOOTP_CLIENTS - 1),
+#undef NB_BOOTP_CLIENTS
+    };
     if ((tapfd = recvfd(sock)) < 0) {
         return tapfd;
     }
@@ -262,19 +352,46 @@ static int parent(int sock, int ready_fd, int exit_fd, const char *api_socket,
     close(sock);
     printf("Starting slirp\n");
     printf("* MTU:             %d\n", cfg->mtu);
-    printf("* Network:         %s\n", inet_ntoa(cfg->vnetwork));
-    printf("* Netmask:         %s\n", inet_ntoa(cfg->vnetmask));
-    printf("* Gateway:         %s\n", inet_ntoa(cfg->vhost));
-    printf("* DNS:             %s\n", inet_ntoa(cfg->vnameserver));
-    printf("* Recommended IP:  %s\n", inet_ntoa(cfg->recommended_vguest));
+    printf("* Network:         %s\n",
+           inet_ntop(AF_INET, &cfg->vnetwork, str, sizeof(str)));
+    printf("* Netmask:         %s\n",
+           inet_ntop(AF_INET, &cfg->vnetmask, str, sizeof(str)));
+    printf("* Gateway:         %s\n",
+           inet_ntop(AF_INET, &cfg->vhost, str, sizeof(str)));
+    printf("* DNS:             %s\n",
+           inet_ntop(AF_INET, &cfg->vnameserver, str, sizeof(str)));
+    printf("* DHCP begin:      %s\n",
+           inet_ntop(AF_INET, &cfg->vdhcp_start, str, sizeof(str)));
+    printf("* DHCP end:        %s\n",
+           inet_ntop(AF_INET, &vdhcp_end, str, sizeof(str)));
+    printf("* Recommended IP:  %s\n",
+           inet_ntop(AF_INET, &cfg->recommended_vguest, str, sizeof(str)));
     if (api_socket != NULL) {
         printf("* API Socket:      %s\n", api_socket);
     }
+#if SLIRP_CONFIG_VERSION_MAX >= 2
+    if (cfg->enable_outbound_addr) {
+        printf(
+            "* Outbound IPv4:    %s\n",
+            inet_ntop(AF_INET, &cfg->outbound_addr.sin_addr, str, sizeof(str)));
+    }
+    if (cfg->enable_outbound_addr6) {
+        if (inet_ntop(AF_INET6, &cfg->outbound_addr6.sin6_addr, str,
+                      sizeof(str)) != NULL) {
+            printf("* Outbound IPv6:    %s\n", str);
+        }
+    }
+#endif
+    if (cfg->vmacaddress_len > 0) {
+        unsigned char *mac = (unsigned char *)cfg->vmacaddress.sa_data;
+        printf("* MAC address:     %02x:%02x:%02x:%02x:%02x:%02x\n", mac[0],
+               mac[1], mac[2], mac[3], mac[4], mac[5]);
+    }
     if (!cfg->disable_host_loopback) {
         printf(
             "WARNING: 127.0.0.1:* on the host is accessible as %s (set "
             "--disable-host-loopback to prohibit connecting to 127.0.0.1:*)\n",
-            inet_ntoa(cfg->vhost));
+            inet_ntop(AF_INET, &cfg->vhost, str, sizeof(str)));
     }
     if (cfg->enable_sandbox && geteuid() != 0) {
         if ((rc = nsenter(target_pid, NULL, NULL, true)) < 0) {
@@ -303,7 +420,7 @@ static int parent(int sock, int ready_fd, int exit_fd, const char *api_socket,
 
 static void usage(const char *argv0)
 {
-    printf("Usage: %s [OPTION]... PID|PATH TAPNAME\n", argv0);
+    printf("Usage: %s [OPTION]... PID|PATH [TAPNAME]\n", argv0);
     printf("User-mode networking for unprivileged network namespaces.\n\n");
     printf("-c, --configure          bring up the interface\n");
     printf("-e, --exit-fd=FD         specify the FD for terminating "
@@ -326,10 +443,29 @@ static void usage(const char *argv0)
            "default=%s)\n",
            DEFAULT_NETNS_TYPE);
     printf("--userns-path=PATH	 specify user namespace path\n");
-    printf("--enable-sandbox         create a new mount namespace and drop all "
-           "capabilities except CAP_NET_BIND_SERVICE (experimental)\n");
+    printf(
+        "--enable-sandbox         create a new mount namespace (and drop all "
+        "caps except CAP_NET_BIND_SERVICE if running as the root)\n");
     printf("--enable-seccomp         enable seccomp to limit syscalls "
            "(experimental)\n");
+    /* v1.1.0 */
+#if SLIRP_CONFIG_VERSION_MAX >= 2
+    printf("--outbound-addr=IPv4     sets outbound ipv4 address to bound to "
+           "(experimental)\n");
+    printf("--outbound-addr6=IPv6    sets outbound ipv6 address to bound to "
+           "(experimental)\n");
+#endif
+#if SLIRP_CONFIG_VERSION_MAX >= 3
+    printf("--disable-dns            disables 10.0.2.3 (or configured internal "
+           "ip) to host dns redirect (experimental)\n");
+#endif
+    /* v1.1.9 */
+    printf("--macaddress=MAC         specify the MAC address of the TAP (only "
+           "valid with -c)\n");
+    /* v1.2.0 */
+    printf("--target-type=TYPE       specify the target type ([netns|bess], "
+           "default=%s)\n",
+           DEFAULT_TARGET_TYPE);
     /* others */
     printf("-h, --help               show this help and exit\n");
     printf("-v, --version            show version and exit\n");
@@ -338,28 +474,42 @@ static void usage(const char *argv0)
 // version output is runc-compatible and machine-parsable
 static void version()
 {
+    const struct scmp_version *scmpv = seccomp_version();
     printf("slirp4netns version %s\n", VERSION ? VERSION : PACKAGE_VERSION);
 #ifdef COMMIT
     printf("commit: %s\n", COMMIT);
 #endif
+    printf("libslirp: %s\n", slirp_version_string());
+    printf("SLIRP_CONFIG_VERSION_MAX: %d\n", SLIRP_CONFIG_VERSION_MAX);
+    if (scmpv != NULL) {
+        printf("libseccomp: %d.%d.%d\n", scmpv->major, scmpv->minor,
+               scmpv->micro);
+        /* Do not free scmpv */
+    }
 }
 
 struct options {
-    pid_t target_pid; // argv[1]
     char *tapname; // argv[2]
-    bool do_config_network; // -c
-    int exit_fd; // -e
-    int ready_fd; // -r
-    unsigned int mtu; // -m
-    bool disable_host_loopback; // --disable-host-loopback
     char *cidr; // --cidr
-    bool enable_ipv6; // -6
     char *api_socket; // -a
     char *netns_type; // argv[1]
     char *netns_path; // --netns-path
     char *userns_path; // --userns-path
+    char *outbound_addr; // --outbound-addr
+    char *outbound_addr6; // --outbound-addr6
+    pid_t target_pid; // argv[1]
+    int exit_fd; // -e
+    int ready_fd; // -r
+    unsigned int mtu; // -m
+    bool do_config_network; // -c
+    bool disable_host_loopback; // --disable-host-loopback
+    bool enable_ipv6; // -6
     bool enable_sandbox; // --enable-sandbox
     bool enable_seccomp; // --enable-seccomp
+    bool disable_dns; // --disable-dns
+    char *macaddress; // --macaddress
+    char *target_type; // --target-type
+    char *bess_socket; // argv[1] (When --target-type="bess")
 };
 
 static void options_init(struct options *options)
@@ -395,6 +545,26 @@ static void options_destroy(struct options *options)
         free(options->userns_path);
         options->userns_path = NULL;
     }
+    if (options->outbound_addr != NULL) {
+        free(options->outbound_addr);
+        options->outbound_addr = NULL;
+    }
+    if (options->outbound_addr6 != NULL) {
+        free(options->outbound_addr6);
+        options->outbound_addr6 = NULL;
+    }
+    if (options->macaddress != NULL) {
+        free(options->macaddress);
+        options->macaddress = NULL;
+    }
+    if (options->target_type != NULL) {
+        free(options->target_type);
+        options->target_type = NULL;
+    }
+    if (options->bess_socket != NULL) {
+        free(options->bess_socket);
+        options->bess_socket = NULL;
+    }
 }
 
 // * caller does not need to call options_init()
@@ -408,12 +578,21 @@ static void parse_args(int argc, char *const argv[], struct options *options)
     char *optarg_netns_type = NULL;
     char *optarg_userns_path = NULL;
     char *optarg_api_socket = NULL;
+    char *optarg_outbound_addr = NULL;
+    char *optarg_outbound_addr6 = NULL;
+    char *optarg_macaddress = NULL;
+    char *optarg_target_type = NULL;
 #define CIDR -42
 #define DISABLE_HOST_LOOPBACK -43
 #define NETNS_TYPE -44
 #define USERNS_PATH -45
 #define ENABLE_SANDBOX -46
 #define ENABLE_SECCOMP -47
+#define OUTBOUND_ADDR -48
+#define OUTBOUND_ADDR6 -49
+#define DISABLE_DNS -50
+#define MACADDRESS -51
+#define TARGET_TYPE -52
 #define _DEPRECATED_NO_HOST_LOOPBACK \
     -10043 // deprecated in favor of disable-host-loopback
 #define _DEPRECATED_CREATE_SANDBOX \
@@ -435,6 +614,11 @@ static void parse_args(int argc, char *const argv[], struct options *options)
         { "enable-seccomp", no_argument, NULL, ENABLE_SECCOMP },
         { "help", no_argument, NULL, 'h' },
         { "version", no_argument, NULL, 'v' },
+        { "outbound-addr", required_argument, NULL, OUTBOUND_ADDR },
+        { "outbound-addr6", required_argument, NULL, OUTBOUND_ADDR6 },
+        { "disable-dns", no_argument, NULL, DISABLE_DNS },
+        { "macaddress", required_argument, NULL, MACADDRESS },
+        { "target-type", required_argument, NULL, TARGET_TYPE },
         { 0, 0, 0, 0 },
     };
     options_init(options);
@@ -491,7 +675,6 @@ static void parse_args(int argc, char *const argv[], struct options *options)
                    "--enable-sandbox instead.\n");
             /* FALLTHROUGH */
         case ENABLE_SANDBOX:
-            printf("WARNING: Support for sandboxing is experimental\n");
             options->enable_sandbox = true;
             break;
         case ENABLE_SECCOMP:
@@ -524,6 +707,23 @@ static void parse_args(int argc, char *const argv[], struct options *options)
             version();
             exit(EXIT_SUCCESS);
             break;
+        case OUTBOUND_ADDR:
+            printf("WARNING: Support for --outbound-addr is experimental\n");
+            optarg_outbound_addr = optarg;
+            break;
+        case OUTBOUND_ADDR6:
+            printf("WARNING: Support for --outbound-addr6 is experimental\n");
+            optarg_outbound_addr6 = optarg;
+            break;
+        case DISABLE_DNS:
+            options->disable_dns = true;
+            break;
+        case MACADDRESS:
+            optarg_macaddress = optarg;
+            break;
+        case TARGET_TYPE:
+            optarg_target_type = optarg;
+            break;
         default:
             goto error;
             break;
@@ -541,6 +741,24 @@ static void parse_args(int argc, char *const argv[], struct options *options)
     if (optarg_api_socket != NULL) {
         options->api_socket = strdup(optarg_api_socket);
     }
+    if (optarg_outbound_addr != NULL) {
+        options->outbound_addr = strdup(optarg_outbound_addr);
+    }
+    if (optarg_outbound_addr6 != NULL) {
+        options->outbound_addr6 = strdup(optarg_outbound_addr6);
+    }
+    if (optarg_macaddress != NULL) {
+        if (!options->do_config_network) {
+            fprintf(stderr, "--macaddr cannot be specified when --configure or "
+                            "-c is not specified\n");
+            goto error;
+        } else {
+            options->macaddress = strdup(optarg_macaddress);
+        }
+    }
+    if (optarg_target_type != NULL) {
+        options->target_type = strdup(optarg_target_type);
+    }
 #undef CIDR
 #undef DISABLE_HOST_LOOPBACK
 #undef NETNS_TYPE
@@ -548,9 +766,53 @@ static void parse_args(int argc, char *const argv[], struct options *options)
 #undef _DEPRECATED_NO_HOST_LOOPBACK
 #undef ENABLE_SANDBOX
 #undef ENABLE_SECCOMP
+#undef OUTBOUND_ADDR
+#undef OUTBOUND_ADDR6
+#undef DISABLE_DNS
+#undef MACADDRESS
+#undef TARGET_TYPE
+
+    /* BESS mode */
+    if (options->target_type != NULL &&
+        strcmp(options->target_type, "bess") == 0) {
+        if (argc - optind < 1) {
+            goto error;
+        }
+        if (argc - optind > 1) {
+            fprintf(stderr, "too many arguments\n");
+            goto error;
+        }
+        options->bess_socket = strdup(argv[optind]);
+        if (options->do_config_network) {
+            fprintf(stderr, "--configure conflicts with --target-type=bess\n");
+            goto error;
+        }
+        if (options->netns_type != NULL) {
+            fprintf(stderr, "--netns-type conflicts with --target-type=bess\n");
+            goto error;
+        }
+        if (options->userns_path != NULL) {
+            fprintf(stderr,
+                    "--userns-path conflicts with --target-type=bess\n");
+            goto error;
+        }
+        printf("WARNING: BESS mode is experimental\n");
+        return;
+    }
+
+    /* NetNS mode*/
+    if (options->target_type != NULL &&
+        strcmp(options->target_type, "netns") != 0) {
+        fprintf(stderr, "--target-type must be either \"netns\" or \"bess\"\n");
+        goto error;
+    }
     if (argc - optind < 2) {
         goto error;
     }
+    if (argc - optind > 2) {
+        // not an error, for preventing potential compatibility issue
+        printf("WARNING: too many arguments\n");
+    }
     if (!options->netns_type ||
         strcmp(options->netns_type, DEFAULT_NETNS_TYPE) == 0) {
         errno = 0;
@@ -668,6 +930,67 @@ finish:
     return rc;
 }
 
+static int get_interface_addr(const char *interface, int af, void *addr)
+{
+    struct ifaddrs *ifaddr, *ifa;
+    if (interface == NULL)
+        return -1;
+
+    if (getifaddrs(&ifaddr) == -1) {
+        fprintf(stderr, "getifaddrs failed to obtain interface addresses");
+        return -1;
+    }
+
+    for (ifa = ifaddr; ifa != NULL; ifa = ifa->ifa_next) {
+        if (ifa->ifa_addr == NULL || ifa->ifa_name == NULL)
+            continue;
+        if (ifa->ifa_addr->sa_family == af) {
+            if (strcmp(ifa->ifa_name, interface) == 0) {
+                if (af == AF_INET) {
+                    *(struct in_addr *)addr =
+                        ((struct sockaddr_in *)ifa->ifa_addr)->sin_addr;
+                } else {
+                    *(struct in6_addr *)addr =
+                        ((struct sockaddr_in6 *)ifa->ifa_addr)->sin6_addr;
+                }
+                return 0;
+            }
+        }
+    }
+    return -1;
+}
+
+/*
+ * Convert a MAC string (macaddr) to bytes (data).
+ * macaddr must be a null-terminated string in the format of
+ * xx:xx:xx:xx:xx:xx.  The data buffer needs to be at least 6 bytes.
+ * Typically the data is put into sockaddr sa_data, which is 14 bytes.
+ */
+static int slirp4netns_macaddr_hexstring_to_data(char *macaddr, char *data)
+{
+    int temp;
+    char *macaddr_ptr;
+    char *data_ptr;
+    for (macaddr_ptr = macaddr, data_ptr = data;
+         macaddr_ptr != NULL && data_ptr < data + 6;
+         macaddr_ptr = strchr(macaddr_ptr, ':'), data_ptr++) {
+        if (macaddr_ptr != macaddr) {
+            macaddr_ptr++; // advance over the :
+        }
+        if (sscanf(macaddr_ptr, "%x", &temp) != 1 || temp < 0 || temp > 255) {
+            fprintf(stderr, "\"%s\" is an invalid MAC address.\n", macaddr);
+            return -1;
+        }
+        *data_ptr = temp;
+    }
+    if (macaddr_ptr != NULL) {
+        fprintf(stderr, "\"%s\" is an invalid MAC address.  Is it too long?\n",
+                macaddr);
+        return -1;
+    }
+    return (int)(data_ptr - data);
+}
+
 static int slirp4netns_config_from_options(struct slirp4netns_config *cfg,
                                            struct options *opt)
 {
@@ -678,10 +1001,90 @@ static int slirp4netns_config_from_options(struct slirp4netns_config *cfg,
     if (rc < 0) {
         goto finish;
     }
-    cfg->enable_ipv6 = cfg->enable_ipv6;
+    cfg->enable_ipv6 = opt->enable_ipv6;
     cfg->disable_host_loopback = opt->disable_host_loopback;
     cfg->enable_sandbox = opt->enable_sandbox;
     cfg->enable_seccomp = opt->enable_seccomp;
+
+#if SLIRP_CONFIG_VERSION_MAX >= 2
+    cfg->enable_outbound_addr = false;
+    cfg->enable_outbound_addr6 = false;
+#endif
+
+    if (opt->outbound_addr != NULL) {
+#if SLIRP_CONFIG_VERSION_MAX >= 2
+        cfg->outbound_addr.sin_family = AF_INET;
+        cfg->outbound_addr.sin_port = 0; // Any local port will do
+        if (inet_pton(AF_INET, opt->outbound_addr,
+                      &cfg->outbound_addr.sin_addr) == 1) {
+            cfg->enable_outbound_addr = true;
+        } else {
+            if (get_interface_addr(opt->outbound_addr, AF_INET,
+                                   &cfg->outbound_addr.sin_addr) != 0) {
+                fprintf(stderr, "outbound-addr has to be valid ipv4 address or "
+                                "interface name.");
+                rc = -1;
+                goto finish;
+            }
+            cfg->enable_outbound_addr = true;
+        }
+#else
+        fprintf(stderr, "slirp4netns has to be compiled against libslrip 4.2.0 "
+                        "or newer for --outbound-addr support.");
+        rc = -1;
+        goto finish;
+#endif
+    }
+    if (opt->outbound_addr6 != NULL) {
+#if SLIRP_CONFIG_VERSION_MAX >= 2
+        cfg->outbound_addr6.sin6_family = AF_INET6;
+        cfg->outbound_addr6.sin6_port = 0; // Any local port will do
+        if (inet_pton(AF_INET6, opt->outbound_addr6,
+                      &cfg->outbound_addr6.sin6_addr) == 1) {
+            cfg->enable_outbound_addr6 = true;
+        } else {
+            if (get_interface_addr(opt->outbound_addr, AF_INET6,
+                                   &cfg->outbound_addr6.sin6_addr) != 0) {
+                fprintf(stderr, "outbound-addr has to be valid ipv4 address or "
+                                "iterface name.");
+                rc = -1;
+                goto finish;
+            }
+            cfg->enable_outbound_addr6 = true;
+        }
+#else
+        fprintf(stderr, "slirp4netns has to be compiled against libslirp 4.2.0 "
+                        "or newer for --outbound-addr6 support.");
+        rc = -1;
+        goto finish;
+#endif
+    }
+
+#if SLIRP_CONFIG_VERSION_MAX >= 3
+    cfg->disable_dns = opt->disable_dns;
+#else
+    if (opt->disable_dns) {
+        fprintf(stderr, "slirp4netns has to be compiled against libslirp 4.3.0 "
+                        "or newer for --disable-dns support.");
+        rc = -1;
+        goto finish;
+    }
+#endif
+
+    cfg->vmacaddress_len = 0;
+    memset(&cfg->vmacaddress, 0, sizeof(cfg->vmacaddress));
+    if (opt->macaddress != NULL) {
+        cfg->vmacaddress.sa_family = AF_LOCAL;
+        int macaddr_len;
+        if ((macaddr_len = slirp4netns_macaddr_hexstring_to_data(
+                 opt->macaddress, cfg->vmacaddress.sa_data)) < 0) {
+            fprintf(stderr, "macaddress has to be a valid MAC address (hex "
+                            "string, 6 bytes, each byte separated by a ':').");
+            rc = -1;
+            goto finish;
+        }
+        cfg->vmacaddress_len = macaddr_len;
+    }
 finish:
     return rc;
 }
@@ -710,9 +1113,16 @@ int main(int argc, char *const argv[])
         goto finish;
     }
     if (child_pid == 0) {
-        if (child(sv[1], options.target_pid, options.do_config_network,
-                  options.tapname, options.netns_path, options.userns_path,
-                  &slirp4netns_config) < 0) {
+        int ret;
+        if (options.target_type != NULL &&
+            strcmp(options.target_type, "bess") == 0) {
+            ret = child_bess(sv[1], options.bess_socket);
+        } else {
+            ret = child(sv[1], options.target_pid, options.do_config_network,
+                        options.tapname, options.netns_path,
+                        options.userns_path, &slirp4netns_config);
+        }
+        if (ret < 0) {
             exit_status = EXIT_FAILURE;
             goto finish;
         }
