@@ -10,14 +10,14 @@ wait_for_network_namespace $child
 
 touch keep_alive
 
-slirp4netns -e 10 $child tun11 10<(while test -e keep_alive; do sleep 0.1; done) &
+slirp4netns -e 10 $child tap11 10<(while test -e keep_alive; do sleep 0.1; done) &
 
 slirp_pid=$!
 
 function cleanup {
-    set +xeuo pipefail
-    kill -9 $child $slirp_pid
-    rm -f keep_alive
+	set +xeuo pipefail
+	kill -9 $child $slirp_pid
+	rm -f keep_alive
 }
 trap cleanup EXIT
 
@@ -29,7 +29,7 @@ rm keep_alive
 wait_process_exits $slirp_pid
 
 if kill -0 $slirp_pid; then
-    exit 1
+	exit 1
 fi
 
 exit 0
